@@ -5,7 +5,7 @@ use warnings;
 use Carp qw/croak/;
 use XSLoader;
 
-our $VERSION = '4.001_001'; # Don't forget to update the TestCompat set for testing against installed encoders!
+our $VERSION = '4.001_002'; # Don't forget to update the TestCompat set for testing against installed encoders!
 our $XS_VERSION = $VERSION; $VERSION= eval $VERSION;
 
 # not for public consumption, just for testing.
@@ -36,6 +36,7 @@ use constant #begin generated
   'SRL_F_DECODER_ALIAS_VARINT' => 8192,
   'SRL_F_DECODER_DECOMPRESS_SNAPPY' => 8,
   'SRL_F_DECODER_DECOMPRESS_ZLIB' => 16,
+  'SRL_F_DECODER_DECOMPRESS_ZSTD' => 131072,
   'SRL_F_DECODER_DESTRUCTIVE_INCREMENTAL' => 1024,
   'SRL_F_DECODER_DIRTY' => 2,
   'SRL_F_DECODER_NEEDS_FINALIZE' => 4,
@@ -45,12 +46,13 @@ use constant #begin generated
   'SRL_F_DECODER_REFUSE_OBJECTS' => 128,
   'SRL_F_DECODER_REFUSE_SNAPPY' => 32,
   'SRL_F_DECODER_REFUSE_ZLIB' => 64,
+  'SRL_F_DECODER_REFUSE_ZSTD' => 262144,
   'SRL_F_DECODER_REUSE' => 1,
   'SRL_F_DECODER_SET_READONLY' => 32768,
   'SRL_F_DECODER_SET_READONLY_SCALARS' => 65536,
   'SRL_F_DECODER_USE_UNDEF' => 16384,
   'SRL_F_DECODER_VALIDATE_UTF8' => 256,
-  'SRL_F_DECODER_VOLATILE_FLAGS' => 2078,
+  'SRL_F_DECODER_VOLATILE_FLAGS' => 133150,
   '_FLAG_NAME' => [
                     'REUSE',
                     'DIRTY',
@@ -68,27 +70,31 @@ use constant #begin generated
                     'ALIAS_VARINT',
                     'USE_UNDEF',
                     'SET_READONLY',
-                    'SET_READONLY_SCALARS'
+                    'SET_READONLY_SCALARS',
+                    'DECOMPRESS_ZSTD',
+                    'REFUSE_ZSTD'
                   ],
   '_FLAG_NAME_STATIC' => [
-                                'REUSE',
-                                undef,
-                                undef,
-                                undef,
-                                undef,
-                                'REFUSE_SNAPPY',
-                                'REFUSE_ZLIB',
-                                'REFUSE_OBJECTS',
-                                'VALIDATE_UTF8',
-                                'NO_BLESS_OBJECTS',
-                                'DESTRUCTIVE_INCREMENTAL',
-                                undef,
-                                'ALIAS_SMALLINT',
-                                'ALIAS_VARINT',
-                                'USE_UNDEF',
-                                'SET_READONLY',
-                                'SET_READONLY_SCALARS'
-                              ],
+                           'REUSE',
+                           undef,
+                           undef,
+                           undef,
+                           undef,
+                           'REFUSE_SNAPPY',
+                           'REFUSE_ZLIB',
+                           'REFUSE_OBJECTS',
+                           'VALIDATE_UTF8',
+                           'NO_BLESS_OBJECTS',
+                           'DESTRUCTIVE_INCREMENTAL',
+                           undef,
+                           'ALIAS_SMALLINT',
+                           'ALIAS_VARINT',
+                           'USE_UNDEF',
+                           'SET_READONLY',
+                           'SET_READONLY_SCALARS',
+                           undef,
+                           'REFUSE_ZSTD'
+                         ],
   '_FLAG_NAME_VOLATILE' => [
                              undef,
                              'DIRTY',
@@ -106,6 +112,8 @@ use constant #begin generated
                              undef,
                              undef,
                              undef,
+                             undef,
+                             'DECOMPRESS_ZSTD',
                              undef
                            ]
 }; #end generated
